import { SdkError } from './errors/SdkError.mjs';
import { mergeBytes } from './utils/bytes.mjs';

/**
 * An object that can serialize and deserialize a value to and from a `Uint8Array`.
 * It supports serializing looser types than it deserializes for convenience.
 * For example, a `bigint` serializer will always deserialize to a `bigint`
 * but can be used to serialize a `number`.
 *
 * @typeParam From - The type of the value to serialize.
 * @typeParam To - The type of the deserialized value. Defaults to `From`.
 */

function mapSerializer(serializer, unmap, map) {
  return {
    description: serializer.description,
    fixedSize: serializer.fixedSize,
    maxSize: serializer.maxSize,
    serialize: value => serializer.serialize(unmap(value)),
    deserialize: (buffer, offset = 0) => {
      const [value, length] = serializer.deserialize(buffer, offset);
      return map ? [map(value, buffer, offset), length] : [value, length];
    }
  };
}

/**
 * Creates a fixed-size serializer from a given serializer.
 *
 * @param serializer - The serializer to wrap into a fixed-size serializer.
 * @param fixedBytes - The fixed number of bytes to read.
 * @param description - A custom description for the serializer.
 */
function fixSerializer(serializer, fixedBytes, description) {
  return {
    description: description ?? `fixed(${fixedBytes}, ${serializer.description})`,
    fixedSize: fixedBytes,
    maxSize: fixedBytes,
    serialize: value => {
      const buffer = new Uint8Array(fixedBytes).fill(0);
      buffer.set(serializer.serialize(value).slice(0, fixedBytes));
      return buffer;
    },
    deserialize: (bytes, offset = 0) => {
      bytes = bytes.slice(offset, offset + fixedBytes);
      if (bytes.length < fixedBytes) {
        throw new SdkError(`Fixed serializer expected ${fixedBytes} bytes, got ${bytes.length}.`);
      }
      const [value] = serializer.deserialize(bytes, 0);
      return [value, offset + fixedBytes];
    }
  };
}
function reverseSerializer(serializer) {
  if (serializer.fixedSize === null) {
    throw new SdkError('Cannot reverse a serializer of variable size.');
  }
  return {
    ...serializer,
    serialize: value => serializer.serialize(value).reverse(),
    deserialize: (bytes, offset = 0) => {
      const fixedSize = serializer.fixedSize;
      const newBytes = mergeBytes([bytes.slice(0, offset), bytes.slice(offset, offset + fixedSize).reverse(), bytes.slice(offset + fixedSize)]);
      return serializer.deserialize(newBytes, offset);
    }
  };
}

export { fixSerializer, mapSerializer, reverseSerializer };
//# sourceMappingURL=Serializer.mjs.map
