import { uniqueSigners, signTransaction } from './Signer.mjs';
import { TRANSACTION_SIZE_LIMIT } from './Transaction.mjs';
import { SdkError } from './errors/SdkError.mjs';

class TransactionBuilder {
  constructor(context, items = [], options = {}) {
    this.context = context;
    this.items = items;
    this.options = options;
  }
  empty() {
    return new TransactionBuilder(this.context, [], this.options);
  }
  setItems(input) {
    return new TransactionBuilder(this.context, this.parseItems(input), this.options);
  }
  prepend(input) {
    return new TransactionBuilder(this.context, [...this.parseItems(input), ...this.items], this.options);
  }
  append(input) {
    return new TransactionBuilder(this.context, [...this.items, ...this.parseItems(input)], this.options);
  }
  add(input) {
    return this.append(input);
  }
  splitAtIndex(index) {
    return [new TransactionBuilder(this.context, this.items.slice(0, index), this.options), new TransactionBuilder(this.context, this.items.slice(index), this.options)];
  }
  setVersion(version) {
    return new TransactionBuilder(this.context, this.items, {
      ...this.options,
      version
    });
  }
  useLegacyVersion() {
    return this.setVersion('legacy');
  }
  useV0() {
    return this.setVersion(0);
  }
  setAddressLookupTables(addressLookupTables) {
    return new TransactionBuilder(this.context, this.items, {
      ...this.options,
      addressLookupTables
    });
  }
  getBlockhash() {
    return typeof this.options.blockhash === 'object' ? this.options.blockhash.blockhash : this.options.blockhash;
  }
  setBlockhash(blockhash) {
    return new TransactionBuilder(this.context, this.items, {
      ...this.options,
      blockhash
    });
  }
  async setLatestBlockhash() {
    return this.setBlockhash(await this.context.rpc.getLatestBlockhash());
  }
  getInstructions() {
    return this.items.map(item => item.instruction);
  }
  getSigners() {
    return uniqueSigners([this.context.payer, ...this.items.flatMap(item => item.signers)]);
  }
  getBytesCreatedOnChain() {
    return this.items.reduce((sum, item) => sum + item.bytesCreatedOnChain, 0);
  }
  async getRentCreatedOnChain() {
    return this.context.rpc.getRent(this.getBytesCreatedOnChain(), {
      includesHeaderBytes: true
    });
  }
  getTransactionSize() {
    return this.context.transactions.serialize(this.setBlockhash('11111111111111111111111111111111').build()).length;
  }
  minimumTransactionsRequired() {
    return Math.ceil(this.getTransactionSize() / TRANSACTION_SIZE_LIMIT);
  }
  fitsInOneTransaction() {
    return this.minimumTransactionsRequired() === 1;
  }
  build() {
    const blockhash = this.getBlockhash();
    if (!blockhash) {
      throw new SdkError('Setting a blockhash is required to build a transaction. ' + 'Please use the `setBlockhash` or `setLatestBlockhash` methods.');
    }
    const input = {
      version: this.options.version ?? 0,
      payer: this.context.payer.publicKey,
      instructions: this.getInstructions(),
      blockhash
    };
    if (input.version === 0 && this.options.addressLookupTables) {
      input.addressLookupTables = this.options.addressLookupTables;
    }
    return this.context.transactions.create(input);
  }
  async buildWithLatestBlockhash() {
    let builder = this;
    if (!this.options.blockhash) {
      builder = await this.setLatestBlockhash();
    }
    return builder.build();
  }
  async buildAndSign() {
    return signTransaction(await this.buildWithLatestBlockhash(), this.getSigners());
  }
  async send(options = {}) {
    const transaction = await this.buildAndSign();
    return this.context.rpc.sendTransaction(transaction, options);
  }
  async sendAndConfirm(options = {}) {
    let builder = this;
    if (!this.options.blockhash) {
      builder = await this.setLatestBlockhash();
    }
    let strategy;
    if (options.confirm?.strategy) {
      strategy = options.confirm.strategy;
    } else {
      const blockhash = typeof builder.options.blockhash === 'object' ? builder.options.blockhash : await builder.context.rpc.getLatestBlockhash();
      strategy = options.confirm?.strategy ?? {
        type: 'blockhash',
        ...blockhash
      };
    }
    const signature = await builder.send(options.send);
    const result = await builder.context.rpc.confirmTransaction(signature, {
      ...options.confirm,
      strategy
    });
    return {
      signature,
      result
    };
  }
  parseItems(input) {
    return (Array.isArray(input) ? input : [input]).flatMap(item => 'instruction' in item ? [item] : item.items);
  }
}
const transactionBuilder = (context, items = []) => new TransactionBuilder(context, items);

export { TransactionBuilder, transactionBuilder };
//# sourceMappingURL=TransactionBuilder.mjs.map
